@@ -9,6 +9,8 @@ our $VERSION;
 use Media::Convert;
 use Carp;
 
+use autodie qw/:all/;
+
 =head1 NAME
 
 Media::Convert::Asset - Media::Convert representation of a media asset
@@ -129,9 +131,52 @@ sub _probe_mtime {
 	return $statdata[9];
 }
 
+=head2 canonical_duration
+
+Which value to take as the canonical duration while probing the duration
+of a media file.
+
+Unfortunately, some versions of ffprobe have bugs in that they either
+read or write corrupt values for some parts of a media file. Sometimes
+the video stream length is correct but the container length is not,
+sometimes it is the other way around.
+
+Since, given that, it is not possible to choose a source that will
+always work as the probed value for the L<duration> attribute, this
+needs to be configurable in some way.
+
+This option can take the following values:
+
+=over 4
+
+=item container
+
+Uses the duration value that is parsed from the container file, rather
+than any particular stream. If the parsing of that value is not buggy,
+this is the right thing to do, and therefore this is also the default.
+
+=item video
+
+Uses the duration value that is parsed from the first video stream.
+
+=item audio
+
+Uses the duration value that is parsed from the first audio stream.
+
+=back
+
+=cut
+
+has 'canonical_duration' => (
+	is => 'ro',
+	isa => 'Str',
+	default => 'container',
+);
+
 =head2 duration
 
-The duration of this asset.
+The duration of this asset. See C<canonical_duration> for details on how
+it is probed.
 
 =cut
 
@@ -147,7 +192,16 @@ sub _probe_duration {
 		return $self->reference->duration;
 	}
 	if($self->duration_style eq "seconds") {
-		return $self->_get_probedata->{format}{duration};
+		if($self->canonical_duration eq 'container') {
+			return $self->_get_probedata->{format}{duration};
+		} elsif($self->canonical_duration eq 'video') {
+			return $self->_get_videodata->{duration};
+		} elsif($self->canonical_duration eq 'audio') {
+			return $self->_get_audiodata->{duration};
+		} else {
+			# don't support this
+			...
+		}
 	} else {
 		return $self->duration_frames;
 	}
@@ -189,6 +243,29 @@ has 'duration_style' => (
 	trigger => \&_warn_duration,
 );
 
+=head2 force_key_frames
+
+Can be set to an expression that can be passed to ffmpeg's C<-force_key_frames>
+parameter. Will not be probed.
+
+=cut
+
+has 'force_key_frames' => (
+	is => 'rw',
+	isa => 'Maybe[Str]',
+	builder => '_probe_force_key_frames',
+	lazy => 1,
+	predicate => 'has_force_key_frames',
+);
+
+sub _probe_force_key_frames {
+	my $self = shift;
+	if($self->has_reference) {
+		return $self->reference->force_key_frames;
+	}
+	return;
+}
+
 =head2 video_codec
 
 The codec in use for the video stream. Note that C<ffprobe> will
@@ -651,24 +728,50 @@ has blackspots => (
 sub _probe_blackspots {
 	my $self = shift;
 	my $blacks = [];
-	pipe R, W;
+	pipe my $R, my $W;
 	if(fork == 0) {
-		open STDERR, ">", "&W";
-		open STDOUT, ">", "&W";
+		open STDERR, ">", "$W";
+		open STDOUT, ">", "$W";
 		my @cmd = ("ffmpeg", "-threads", "1", "-nostats", "-i", $self->url, "-vf", "blackdetect=d=0:pix_th=.01", "-f", "null", "/dev/null");
 		exec @cmd;
 		die "exec failed";
 	}
-	close W;
-	while(<R>) {
+	close $W;
+	while(<$R>) {
 		if(/blackdetect.*black_start:(?<start>[\d\.]+)\sblack_end:(?<end>[\d\.]+)\sblack_duration:(?<duration>[\d\.]+)/) {
 			push @$blacks, { %+ };
 		}
 	}
-	close(R);
+	close($R);
 	return $blacks;
 }
 
+
+=head2 channel_layouts
+
+Returns an array of audio channel layouts, as detected by ffprobe.
+
+=cut
+
+has 'channel_layouts' => (
+	is => 'rw',
+	traits => ['Array'],
+	isa => 'ArrayRef[Str]',
+	builder => '_probe_channel_layouts',
+	lazy => 1,
+);
+
+sub _probe_channel_layouts {
+	my $self = shift;
+	my $rv = [];
+	foreach my $stream(@{$self->_get_probedata->{streams}}) {
+		if($stream->{codec_type} eq "audio") {
+			push @$rv, $stream->{channel_layout};
+		}
+	}
+	return $rv;
+}
+
 =head2 astream_ids
 
 Returns an array with the IDs for the audio streams in this file.
@@ -692,13 +795,13 @@ has 'astream_ids' => (
 
 sub _probe_astream_ids {
 	my $self = shift;
-	my @rv;
+	my $rv = [];
 	foreach my $stream(@{$self->_get_probedata->{streams}}) {
 		if($stream->{codec_type} eq "audio") {
-			push @rv, $stream->{index};
+			push @$rv, $stream->{index};
 		}
 	}
-	return \@rv;
+	return $rv;
 }
 
 =head2 audio_channel_count
@@ -742,9 +845,23 @@ sub _probe_vstream_id {
 
 =head2 extra_params
 
-Add extra parameters. This should be used sparingly, rather add some
+Add extra (output) parameters. This should be used sparingly, rather add some
 abstraction.
 
+=head3 handles
+
+=over
+
+=item *
+
+add_param (adds an extra parameter)
+
+=item *
+
+drop_param (delete a parameter)
+
+=back
+
 =cut
 
 has 'extra_params' => (
@@ -767,6 +884,47 @@ sub _probe_extra_params {
 	return {};
 }
 
+=head2 input_params
+
+Add extra input parameters. This should be used sparingly, rather add
+some abstraction.
+
+=head3 handles
+
+=over
+
+=item *
+
+add_input_param (adds an extra parameter)
+
+=item *
+
+drop_input_param (delete a parameter)
+
+=back
+
+=cut
+
+has 'input_params' => (
+	traits => ['Hash'],
+	isa => 'HashRef[Str]',
+	is => 'ro',
+	handles => {
+		add_input_param => 'set',
+		drop_input_param => 'delete',
+	},
+	builder => "_probe_input_params",
+	lazy => 1,
+);
+
+sub _probe_input_params {
+	my $self = shift;
+	if($self->has_reference) {
+		return $self->reference->input_params;
+	}
+	return {};
+}
+
 =head2 time_offset
 
 Apply an input time offset to this video (only valid when used as an
@@ -816,6 +974,13 @@ sub readopts {
 	if($self->has_time_offset) {
 		push @opts, ("-itsoffset", $self->time_offset);
 	}
+
+	if(scalar(keys(%{$self->input_params})) > 0) {
+		foreach my $param(keys %{$self->input_params}) {
+			push @opts, ("-$param", $self->input_params->{$param});
+		}
+	}
+
 	push @opts, ("-i", $self->url);
 	return @opts;
 }
@@ -845,6 +1010,9 @@ sub writeopts {
 		if(defined($self->video_preset)) {
 			push @opts, ('-preset', $self->video_preset);
 		}
+		if(defined($self->force_key_frames)) {
+			push @opts, ('-force_key_frames', $self->force_key_frames);
+		}
 		if($self->has_pass) {
 			push @opts, ('-pass', $self->pass, '-passlogfile', $self->url . '-multipass');
 		}
@@ -906,13 +1074,16 @@ sub _probe {
 	if($self->has_reference) {
 		return $self->reference->_get_probedata;
 	}
-	open my $jsonpipe, "-|:encoding(UTF-8)", "ffprobe", "-loglevel", "quiet", "-print_format", "json", "-show_format", "-show_streams", $self->url;
-	my $json = "";
-	while(<$jsonpipe>) {
-		$json .= $_;
+	if(-e $self->url) {
+		open my $jsonpipe, "-|:encoding(UTF-8)", "ffprobe", "-loglevel", "quiet", "-print_format", "json", "-show_format", "-show_streams", $self->url;
+		my $json = "";
+		while(<$jsonpipe>) {
+			$json .= $_;
+		}
+		close $jsonpipe;
+		return decode_json($json);
 	}
-	close $jsonpipe;
-	return decode_json($json);
+	return {};
 }
 
 sub _probe_audiodata {
